@@ -390,7 +390,7 @@ fn should_forward_by_response(
             }
             macro_rules! examine_record {
                 ($rec:ident, $is_answer:expr) => {
-                    if let RData::CNAME(ref name) = $rec.rdata() {
+                    if let Some(RData::CNAME(name)) = $rec.data() {
                         if $is_answer {
                             if let Some(value) = check_name_in_proxy_list(acl, name) {
                                 return value;
@@ -407,14 +407,14 @@ fn should_forward_by_response(
                         );
                         return true;
                     }
-                    let forward = match $rec.rdata() {
-                        RData::A(ref ip) => acl.check_ip_in_proxy_list(&IpAddr::V4(*ip)),
-                        RData::AAAA(ref ip) => acl.check_ip_in_proxy_list(&IpAddr::V6(*ip)),
+                    let forward = match $rec.data() {
+                        Some(RData::A(ip)) => acl.check_ip_in_proxy_list(&IpAddr::V4(*ip)),
+                        Some(RData::AAAA(ip)) => acl.check_ip_in_proxy_list(&IpAddr::V6(*ip)),
                         // MX records cause type A additional section processing for the host specified by EXCHANGE.
-                        RData::MX(ref mx) => examine_name!(mx.exchange(), $is_answer),
+                        Some(RData::MX(mx)) => examine_name!(mx.exchange(), $is_answer),
                         // NS records cause both the usual additional section processing to locate a type A record...
-                        RData::NS(ref name) => examine_name!(name, $is_answer),
-                        RData::PTR(_) => unreachable!(),
+                        Some(RData::NS(name)) => examine_name!(name, $is_answer),
+                        Some(RData::PTR(_)) => unreachable!(),
                         _ => acl.is_default_in_proxy_list(),
                     };
                     if !forward {
@@ -492,9 +492,9 @@ impl DnsClient {
             if let Ok(result) = r {
                 for rec in result.answers() {
                     trace!("dns answer: {:?}", rec);
-                    match *rec.rdata() {
-                        RData::A(ip) => self.context.add_to_reverse_lookup_cache(ip.into(), forward).await,
-                        RData::AAAA(ip) => self.context.add_to_reverse_lookup_cache(ip.into(), forward).await,
+                    match rec.data() {
+                        Some(RData::A(ip)) => self.context.add_to_reverse_lookup_cache((*ip).into(), forward).await,
+                        Some(RData::AAAA(ip)) => self.context.add_to_reverse_lookup_cache((*ip).into(), forward).await,
                         _ => (),
                     }
                 }
