@@ -4,27 +4,29 @@ use std::{
     mem,
     net::{IpAddr, SocketAddr},
     pin::Pin,
-    sync::Arc,
+    sync::{
+        atomic::{AtomicBool, Ordering},
+        Arc,
+    },
     task::{Context, Poll, Waker},
-    time::Duration as StdDuration,
+    thread::{self, JoinHandle, Thread},
+    time::Duration,
 };
 
 use log::{error, trace};
-use parking_lot::Mutex as ParkingMutex;
 use shadowsocks::{net::TcpSocketOpts, relay::socks5::Address};
 use smoltcp::{
     iface::{Interface, InterfaceBuilder, Routes, SocketHandle},
     phy::{DeviceCapabilities, Medium},
     socket::{TcpSocket, TcpSocketBuffer, TcpState},
     storage::RingBuffer,
-    time::{Duration, Instant},
+    time::{Duration as SmolDuration, Instant as SmolInstant},
     wire::{IpAddress, IpCidr, Ipv4Address, Ipv6Address, TcpPacket},
 };
+use spin::Mutex as SpinMutex;
 use tokio::{
     io::{AsyncRead, AsyncWrite, ReadBuf},
-    sync::{mpsc, Notify},
-    task::JoinHandle,
-    time,
+    sync::mpsc,
 };
 
 use crate::local::{
@@ -36,11 +38,9 @@ use crate::local::{
 
 use super::virt_device::VirtTunDevice;
 
-// NOTE: Default value is taken from Linux
-// recv: /proc/sys/net/ipv4/tcp_rmem 87380 bytes
-// send: /proc/sys/net/ipv4/tcp_wmem 16384 bytes
-const DEFAULT_TCP_SEND_BUFFER_SIZE: u32 = 16384;
-const DEFAULT_TCP_RECV_BUFFER_SIZE: u32 = 87380;
+// NOTE: Default buffer could contain 20 AEAD packets
+const DEFAULT_TCP_SEND_BUFFER_SIZE: u32 = 0x3FFF * 20;
+const DEFAULT_TCP_RECV_BUFFER_SIZE: u32 = 0x3FFF * 20;
 
 struct TcpSocketControl {
     send_buffer: RingBuffer<'static, u8>,
@@ -50,17 +50,36 @@ struct TcpSocketControl {
     is_closed: bool,
 }
 
-struct TcpSocketManager {
-    iface: Interface<'static, VirtTunDevice>,
-    manager_notify: Arc<Notify>,
-    sockets: HashMap<SocketHandle, Arc<ParkingMutex<TcpSocketControl>>>,
+struct ManagerNotify {
+    thread: Thread,
 }
 
-type SharedTcpSocketManager = Arc<ParkingMutex<TcpSocketManager>>;
+impl ManagerNotify {
+    fn new(thread: Thread) -> ManagerNotify {
+        ManagerNotify { thread }
+    }
+
+    fn notify(&self) {
+        self.thread.unpark();
+    }
+}
+
+struct TcpSocketManager {
+    iface: Interface<'static, VirtTunDevice>,
+    sockets: HashMap<SocketHandle, SharedTcpConnectionControl>,
+    socket_creation_rx: mpsc::UnboundedReceiver<TcpSocketCreation>,
+}
+
+type SharedTcpConnectionControl = Arc<SpinMutex<TcpSocketControl>>;
+
+struct TcpSocketCreation {
+    control: SharedTcpConnectionControl,
+    socket: TcpSocket<'static>,
+}
 
 struct TcpConnection {
-    control: Arc<ParkingMutex<TcpSocketControl>>,
-    manager_notify: Arc<Notify>,
+    control: SharedTcpConnectionControl,
+    manager_notify: Arc<ManagerNotify>,
 }
 
 impl Drop for TcpConnection {
@@ -71,25 +90,27 @@ impl Drop for TcpConnection {
 }
 
 impl TcpConnection {
-    fn new(socket: TcpSocket<'static>, manager: SharedTcpSocketManager, tcp_opts: &TcpSocketOpts) -> TcpConnection {
+    fn new(
+        socket: TcpSocket<'static>,
+        socket_creation_tx: &mpsc::UnboundedSender<TcpSocketCreation>,
+        manager_notify: Arc<ManagerNotify>,
+        tcp_opts: &TcpSocketOpts,
+    ) -> TcpConnection {
         let send_buffer_size = tcp_opts.send_buffer_size.unwrap_or(DEFAULT_TCP_SEND_BUFFER_SIZE);
         let recv_buffer_size = tcp_opts.recv_buffer_size.unwrap_or(DEFAULT_TCP_RECV_BUFFER_SIZE);
 
-        let (control, manager_notify) = {
-            let mut manager = manager.lock();
-            let socket_handle = manager.iface.add_socket(socket);
+        let control = Arc::new(SpinMutex::new(TcpSocketControl {
+            send_buffer: RingBuffer::new(vec![0u8; send_buffer_size as usize]),
+            send_waker: None,
+            recv_buffer: RingBuffer::new(vec![0u8; recv_buffer_size as usize]),
+            recv_waker: None,
+            is_closed: false,
+        }));
 
-            let control = Arc::new(ParkingMutex::new(TcpSocketControl {
-                send_buffer: RingBuffer::new(vec![0u8; send_buffer_size as usize]),
-                send_waker: None,
-                recv_buffer: RingBuffer::new(vec![0u8; recv_buffer_size as usize]),
-                recv_waker: None,
-                is_closed: false,
-            }));
-
-            manager.sockets.insert(socket_handle.clone(), control.clone());
-            (control, manager.manager_notify.clone())
-        };
+        let _ = socket_creation_tx.send(TcpSocketCreation {
+            control: control.clone(),
+            socket,
+        });
 
         TcpConnection {
             control,
@@ -124,7 +145,9 @@ impl AsyncRead for TcpConnection {
         let n = control.recv_buffer.dequeue_slice(recv_buf);
         buf.advance(n);
 
-        self.manager_notify.notify_one();
+        if control.recv_buffer.is_empty() {
+            self.manager_notify.notify();
+        }
         Ok(()).into()
     }
 }
@@ -150,7 +173,9 @@ impl AsyncWrite for TcpConnection {
 
         let n = control.send_buffer.enqueue_slice(buf);
 
-        self.manager_notify.notify_one();
+        if control.send_buffer.is_full() {
+            self.manager_notify.notify();
+        }
         Ok(n).into()
     }
 
@@ -178,17 +203,19 @@ impl AsyncWrite for TcpConnection {
 
 pub struct TcpTun {
     context: Arc<ServiceContext>,
-    manager: SharedTcpSocketManager,
-    manager_handle: JoinHandle<()>,
-    manager_notify: Arc<Notify>,
+    manager_handle: Option<JoinHandle<()>>,
+    manager_notify: Arc<ManagerNotify>,
+    manager_socket_creation_tx: mpsc::UnboundedSender<TcpSocketCreation>,
+    manager_running: Arc<AtomicBool>,
     balancer: PingBalancer,
     iface_rx: mpsc::UnboundedReceiver<Vec<u8>>,
-    iface_tx: mpsc::Sender<Vec<u8>>,
+    iface_tx: mpsc::UnboundedSender<Vec<u8>>,
 }
 
 impl Drop for TcpTun {
     fn drop(&mut self) {
-        self.manager_handle.abort();
+        self.manager_running.store(false, Ordering::Relaxed);
+        let _ = self.manager_handle.take().unwrap().join();
     }
 }
 
@@ -218,151 +245,156 @@ impl TcpTun {
             .routes(iface_routes)
             .finalize();
 
-        let manager_notify = Arc::new(Notify::new());
-        let manager = Arc::new(ParkingMutex::new(TcpSocketManager {
+        let (manager_socket_creation_tx, manager_socket_creation_rx) = mpsc::unbounded_channel();
+        let mut manager = TcpSocketManager {
             iface,
-            manager_notify: manager_notify.clone(),
             sockets: HashMap::new(),
-        }));
+            socket_creation_rx: manager_socket_creation_rx,
+        };
+
+        let manager_running = Arc::new(AtomicBool::new(true));
 
         let manager_handle = {
-            let manager = manager.clone();
-            let manager_notify = manager_notify.clone();
-            tokio::spawn(async move {
-                loop {
-                    let next_duration = {
-                        let TcpSocketManager {
-                            ref mut iface,
-                            ref mut sockets,
-                            ..
-                        } = *(manager.lock());
+            let manager_running = manager_running.clone();
 
-                        let before_poll = Instant::now();
-                        let updated_sockets = match iface.poll(before_poll) {
-                            Ok(u) => u,
-                            Err(err) => {
-                                error!("VirtDevice::poll error: {}", err);
-                                false
-                            }
-                        };
+            thread::spawn(move || {
+                let TcpSocketManager {
+                    ref mut iface,
+                    ref mut sockets,
+                    ref mut socket_creation_rx,
+                    ..
+                } = manager;
 
-                        let after_poll = Instant::now();
+                while manager_running.load(Ordering::Relaxed) {
+                    while let Ok(TcpSocketCreation { control, socket }) = socket_creation_rx.try_recv() {
+                        let handle = iface.add_socket(socket);
+                        sockets.insert(handle, control);
+                    }
 
-                        if updated_sockets {
-                            trace!("VirtDevice::poll costed {}", after_poll - before_poll);
+                    let before_poll = SmolInstant::now();
+                    let updated_sockets = match iface.poll(before_poll) {
+                        Ok(u) => u,
+                        Err(err) => {
+                            error!("VirtDevice::poll error: {}", err);
+                            false
                         }
-
-                        // Check all the sockets' status
-                        let mut sockets_to_remove = Vec::new();
-
-                        for (socket_handle, control) in sockets.iter() {
-                            let socket_handle = socket_handle.clone();
-                            let socket = iface.get_socket::<TcpSocket>(socket_handle);
-                            let mut control = control.lock();
-
-                            #[inline]
-                            fn close_socket_control(control: &mut TcpSocketControl) {
-                                control.is_closed = true;
-                                if let Some(waker) = control.send_waker.take() {
-                                    waker.wake();
-                                }
-                                if let Some(waker) = control.recv_waker.take() {
-                                    waker.wake();
-                                }
-                            }
-
-                            if !socket.is_open() || socket.state() == TcpState::Closed {
-                                sockets_to_remove.push(socket_handle);
-                                close_socket_control(&mut *control);
-                                continue;
-                            }
-
-                            if control.is_closed {
-                                // Close the socket.
-                                socket.close();
-                                // sockets_to_remove.push(socket_handle);
-                                // close_socket_control(&mut *control);
-                                continue;
-                            }
-
-                            // Check if readable
-                            let mut has_received = false;
-                            while socket.can_recv() && !control.recv_buffer.is_full() {
-                                let result = socket.recv(|buffer| {
-                                    let n = control.recv_buffer.enqueue_slice(buffer);
-                                    (n, ())
-                                });
-
-                                match result {
-                                    Ok(..) => {
-                                        has_received = true;
-                                    }
-                                    Err(err) => {
-                                        error!("socket recv error: {}", err);
-                                        sockets_to_remove.push(socket_handle);
-                                        close_socket_control(&mut *control);
-                                        break;
-                                    }
-                                }
-                            }
-
-                            if has_received {
-                                if let Some(waker) = control.recv_waker.take() {
-                                    waker.wake();
-                                }
-                            }
-
-                            // Check if writable
-                            let mut has_sent = false;
-                            while socket.can_send() && !control.send_buffer.is_empty() {
-                                let result = socket.send(|buffer| {
-                                    let n = control.send_buffer.dequeue_slice(buffer);
-                                    (n, ())
-                                });
-
-                                match result {
-                                    Ok(..) => {
-                                        has_sent = true;
-                                    }
-                                    Err(err) => {
-                                        error!("socket send error: {}", err);
-                                        sockets_to_remove.push(socket_handle);
-                                        close_socket_control(&mut *control);
-                                        break;
-                                    }
-                                }
-                            }
-
-                            if has_sent {
-                                if let Some(waker) = control.send_waker.take() {
-                                    waker.wake();
-                                }
-                            }
-                        }
-
-                        for socket_handle in sockets_to_remove {
-                            sockets.remove(&socket_handle);
-                            iface.remove_socket(socket_handle);
-                        }
-
-                        let next_duration = iface.poll_delay(Instant::now()).unwrap_or(Duration::from_millis(50));
-
-                        next_duration
                     };
 
-                    tokio::select! {
-                        _ = time::sleep(StdDuration::from(next_duration)) => {}
-                        _ = manager_notify.notified() => {}
+                    if updated_sockets {
+                        trace!("VirtDevice::poll costed {}", SmolInstant::now() - before_poll);
+                    }
+
+                    // Check all the sockets' status
+                    let mut sockets_to_remove = Vec::new();
+
+                    for (socket_handle, control) in sockets.iter() {
+                        let socket_handle = socket_handle.clone();
+                        let socket = iface.get_socket::<TcpSocket>(socket_handle);
+                        let mut control = control.lock();
+
+                        #[inline]
+                        fn close_socket_control(control: &mut TcpSocketControl) {
+                            control.is_closed = true;
+                            if let Some(waker) = control.send_waker.take() {
+                                waker.wake();
+                            }
+                            if let Some(waker) = control.recv_waker.take() {
+                                waker.wake();
+                            }
+                        }
+
+                        if !socket.is_open() || socket.state() == TcpState::Closed {
+                            sockets_to_remove.push(socket_handle);
+                            close_socket_control(&mut *control);
+                            continue;
+                        }
+
+                        if control.is_closed {
+                            // Close the socket.
+                            socket.close();
+                            // sockets_to_remove.push(socket_handle);
+                            // close_socket_control(&mut *control);
+                            continue;
+                        }
+
+                        // Check if readable
+                        let mut has_received = false;
+                        while socket.can_recv() && !control.recv_buffer.is_full() {
+                            let result = socket.recv(|buffer| {
+                                let n = control.recv_buffer.enqueue_slice(buffer);
+                                (n, ())
+                            });
+
+                            match result {
+                                Ok(..) => {
+                                    has_received = true;
+                                }
+                                Err(err) => {
+                                    error!("socket recv error: {}", err);
+                                    sockets_to_remove.push(socket_handle);
+                                    close_socket_control(&mut *control);
+                                    break;
+                                }
+                            }
+                        }
+
+                        if has_received && control.recv_waker.is_some() {
+                            if let Some(waker) = control.recv_waker.take() {
+                                waker.wake();
+                            }
+                        }
+
+                        // Check if writable
+                        let mut has_sent = false;
+                        while socket.can_send() && !control.send_buffer.is_empty() {
+                            let result = socket.send(|buffer| {
+                                let n = control.send_buffer.dequeue_slice(buffer);
+                                (n, ())
+                            });
+
+                            match result {
+                                Ok(..) => {
+                                    has_sent = true;
+                                }
+                                Err(err) => {
+                                    error!("socket send error: {}", err);
+                                    sockets_to_remove.push(socket_handle);
+                                    close_socket_control(&mut *control);
+                                    break;
+                                }
+                            }
+                        }
+
+                        if has_sent && control.send_waker.is_some() {
+                            if let Some(waker) = control.send_waker.take() {
+                                waker.wake();
+                            }
+                        }
+                    }
+
+                    for socket_handle in sockets_to_remove {
+                        sockets.remove(&socket_handle);
+                        iface.remove_socket(socket_handle);
+                    }
+
+                    let next_duration = iface.poll_delay(before_poll).unwrap_or(SmolDuration::from_millis(5));
+                    if next_duration != SmolDuration::ZERO {
+                        thread::park_timeout(Duration::from(next_duration));
                     }
                 }
+
+                trace!("VirtDevice::poll thread exited");
             })
         };
 
+        let manager_notify = Arc::new(ManagerNotify::new(manager_handle.thread().clone()));
+
         TcpTun {
             context,
-            manager,
-            manager_handle,
+            manager_handle: Some(manager_handle),
             manager_notify,
+            manager_socket_creation_tx,
+            manager_running,
             balancer,
             iface_rx,
             iface_tx,
@@ -387,8 +419,10 @@ impl TcpTun {
                 TcpSocketBuffer::new(vec![0u8; send_buffer_size as usize]),
             );
             socket.set_keep_alive(accept_opts.tcp.keepalive.map(From::from));
-            // FIXME: This should follows system's setting. 7200 is Linux's default.
-            socket.set_timeout(Some(Duration::from_secs(7200)));
+            // FIXME: It should follow system's setting. 7200 is Linux's default.
+            socket.set_timeout(Some(SmolDuration::from_secs(7200)));
+            // NO ACK delay
+            // socket.set_ack_delay(None);
 
             if let Err(err) = socket.listen(dst_addr) {
                 return Err(io::Error::new(ErrorKind::Other, err));
@@ -396,7 +430,12 @@ impl TcpTun {
 
             trace!("created TCP connection for {} <-> {}", src_addr, dst_addr);
 
-            let connection = TcpConnection::new(socket, self.manager.clone(), &accept_opts.tcp);
+            let connection = TcpConnection::new(
+                socket,
+                &self.manager_socket_creation_tx,
+                self.manager_notify.clone(),
+                &accept_opts.tcp,
+            );
 
             // establish a tunnel
             let context = self.context.clone();
@@ -412,12 +451,12 @@ impl TcpTun {
     }
 
     pub async fn drive_interface_state(&mut self, frame: &[u8]) {
-        if let Err(..) = self.iface_tx.send(frame.to_vec()).await {
+        if let Err(..) = self.iface_tx.send(frame.to_vec()) {
             panic!("interface send channel closed unexpectly");
         }
 
         // Wake up and poll the interface.
-        self.manager_notify.notify_one();
+        self.manager_notify.notify();
     }
 
     pub async fn recv_packet(&mut self) -> Vec<u8> {
